// card.c

// 3-10: Now thread-safe(r) with help from Jon Watte (hplus@b500.com)

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <string.h>
#include "radiocard.h"

__declspec(dllexport)
// Implemented to comply with Be driver interface; they don't do much
status_t init_hardware(void);	// Mutes card
status_t init_driver(void);
void uninit_driver(void);

//status_t r_open( const char *, uint32, void** );

const char ** publish_devices(void);
device_hooks * find_device(const char *);

///////////////////////////////////////////////
// Read from card
unsigned short read_card(int sub_port);
// Write to card
void write_card(int sub_port, unsigned short value);
// Mute card
void CardOff(void);
// Tune to specific frequency
void Tune(float freq);
// I/O Helper functions
void zero( void );
void one( void );

// Thanks to Jon Watte for pointing out some (big) deficiencies
sem_id g_sem;

/* init_hardware - called once the first time the driver is loaded */
status_t
init_hardware (void)
{
	CardOff();	// Mute on startup
	return B_OK;
}

status_t
init_driver (void)
{
	g_sem = create_sem( 1, "NewsRadio lock" );
	return (g_sem < 0) ? g_sem : B_OK;
}

void
uninit_driver (void)
{
//	CardOff();
	delete_sem( g_sem );
}

static status_t
r_open (const char *name, uint32 flags, void** cookie)
{
	return B_OK;
}

static status_t
r_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	*num_bytes = 0;				/* tell caller nothing was read */
	return B_IO_ERROR;
}


/* r_write: handle write() calls.
	3/10/99: Functionality temporarily removed because it didn't work right */
static status_t
r_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	return B_OK;
}


/* r_control: decides whether to tune or mute the card */
static status_t
r_control (void* cookie, uint32 op, void* arg, size_t len)
{
	switch(op) {
	case RADIO_TUNE:
		Tune(*(float *)arg);
		break;
	case RADIO_OFF:
		CardOff();
		break;
	default:
		return EINVAL;
	}
	return B_OK;
}

static status_t
r_close (void* cookie)
{
	return B_OK;
}

static status_t
r_free (void* cookie)
{
	return B_OK;
}

/* null-terminated array of device names */
static const char *my_device_name[] = {
	"misc/radio",
	NULL
};

device_hooks my_device_hooks = {
	r_open, 			/* -> open entry point */
	r_close, 			/* -> close entry point */
	r_free,				/* -> free cookie */
	r_control, 			/* -> control entry point */
	r_read,				/* -> read entry point */
	r_write				/* -> write entry point */
};

/* publish_devices - return a null-terminated array of devices
	supported by this driver */
const char**
publish_devices()
{
	return my_device_name;
}

/* find_device - return ptr to device hooks structure */
device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

/* reads a byte in from the card */
unsigned short read_card(int sub_port)
{
   return read_io_8(RADIO_BASE+sub_port);
}

/* writes a byte to the card */
void write_card(int sub_port, unsigned short value)
{
   write_io_8(RADIO_BASE+sub_port,value);
}

/* switch card off */
/* 3-10: Now thread-safe(r) */
void CardOff(void)
{
	int err;
	err = acquire_sem( g_sem );
	
	if ( err >= B_OK ) {	// No trouble encountered
	   write_card(0,1);
	   release_sem( g_sem );
	}
}

/* Tune to frequency "freq" */
/* 3-10: Now thread-safe(r) */
void Tune( float freq )
{
	int i, err;
	double TDFreq = (double)freq * 16000;
	unsigned long TmpFreq = (long) TDFreq;
	
	err = acquire_sem( g_sem );
	
	if ( err >= B_OK ) {	// No trouble encountered
		TmpFreq = TmpFreq / 200 + 856;
		write_card( 0,0xc8 );
		write_card( 0,0xc9 );
		write_card( 0,0xc9 );
		
		for (i = 0; i < 10; i++)
			zero();
			
		for (i = 14; i >= 0; i--)
			if (TmpFreq & (1 << i))
				one();
			else
				zero();
			
		write_card( 0,0xc8);
	
		write_card( 0,0 );
		release_sem( g_sem );
	}
}	

// I/O helper functions
void zero( void )
{
	write_card( 0,1 );
	write_card( 0,3 );
	write_card( 0,1 );
}

void one( void )
{
	write_card( 0,5 );
	write_card( 0,7 );
	write_card( 0,5 );
}

